"""Pytest configuration and fixtures for smlight_cc_flasher tests."""

import pytest

from smlight_cc_flasher.gpio import GpioConfig, GpioPattern


@pytest.fixture
def mock_gpiod(mocker):
    """Mock gpiod module."""
    mocker.patch("smlight_cc_flasher.gpio.os.path.exists", return_value=True)

    mock = mocker.patch("smlight_cc_flasher.gpio.gpiod")

    mock.LineSettings.return_value = mocker.MagicMock()

    mock_request = mocker.MagicMock()
    mock.request_lines.return_value.__enter__.return_value = mock_request
    mock.request_lines.return_value.__exit__.return_value = None

    mock.line.Direction.OUTPUT = "OUTPUT"
    mock.line.Value.return_value = mocker.MagicMock()

    return mock


@pytest.fixture
def sample_gpio_config():
    """Sample GPIO configuration for testing."""
    return GpioConfig(
        chip="gpiochip0",
        patterns=[
            GpioPattern(pins={11: True, 12: False}, delay_after=0.1),
            GpioPattern(pins={11: False, 12: False}, delay_after=0.1),
            GpioPattern(pins={11: True, 12: True}, delay_after=0.1),
        ],
    )



"""Tests for GPIO module."""

import pytest

from smlight_cc_flasher.gpio import GpioPattern, send_gpio_pattern


def test_gpio_pattern_dataclass():
    """Test GpioPattern dataclass creation."""
    pattern = GpioPattern(pins={11: True, 12: False}, delay_after=0.5)

    assert pattern.pins == {11: True, 12: False}
    assert pattern.delay_after == 0.5


def test_gpio_config_dataclass(sample_gpio_config):
    """Test GpioConfig dataclass creation."""
    assert sample_gpio_config.chip == "gpiochip0"
    assert len(sample_gpio_config.patterns) == 3
    assert sample_gpio_config.patterns[0].pins == {11: True, 12: False}
    assert sample_gpio_config.patterns[1].pins == {11: False, 12: False}
    assert sample_gpio_config.patterns[2].pins == {11: True, 12: True}
    assert all(p.delay_after == 0.1 for p in sample_gpio_config.patterns)


@pytest.mark.asyncio
async def test_send_gpio_pattern_basic(mock_gpiod, mocker, sample_gpio_config):
    """Test basic GPIO pattern sending."""
    mocker.patch("smlight_cc_flasher.gpio.time.sleep")

    await send_gpio_pattern(sample_gpio_config.chip, sample_gpio_config.patterns)

    mock_gpiod.request_lines.assert_called_once()
    call_args = mock_gpiod.request_lines.call_args
    assert call_args[1]["path"] == f"/dev/{sample_gpio_config.chip}"
    assert call_args[1]["consumer"] == "smlight-cc-flasher"

    mock_request = mock_gpiod.request_lines.return_value.__enter__.return_value
    assert mock_request.set_values.call_count == len(sample_gpio_config.patterns) - 1



import asyncio
import dataclasses
import logging
import os
import time

import gpiod

logger = logging.getLogger(__name__)


def _resolve_chip_path(chip: str) -> str:
    """Resolve chip name to full device path."""
    if chip.startswith("/dev"):
        return chip

    chip_path = f"/dev/{chip}"
    if not os.path.exists(chip_path):
        raise FileNotFoundError(f"GPIO chip '{chip}' not found at {chip_path}. ")
    return chip_path


@dataclasses.dataclass
class GpioPattern:
    pins: dict[int, bool]
    delay_after: float


@dataclasses.dataclass
class GpioConfig:
    chip: str
    patterns: list[GpioPattern]


gpioResets = {
    "smhub": GpioConfig(
        chip="gpiochip2",
        patterns=[
            GpioPattern(pins={11: True, 12: True}, delay_after=0.1),
            GpioPattern(pins={11: False, 12: False}, delay_after=0.1),
            GpioPattern(pins={11: True, 12: False}, delay_after=0.1),
            GpioPattern(pins={11: True, 12: True}, delay_after=0.1),
        ],
    ),
    "smhub-1": GpioConfig(
        chip="gpiochip1",
        patterns=[
            GpioPattern(pins={11: True, 12: True}, delay_after=0.1),
            GpioPattern(pins={11: False, 12: False}, delay_after=0.1),
            GpioPattern(pins={11: True, 12: False}, delay_after=0.1),
            GpioPattern(pins={11: True, 12: True}, delay_after=0.1),
        ],
    ),
}


def _send_gpio_pattern(chip: str, pattern: list[GpioPattern]) -> None:
    """Send GPIO pattern to chip."""
    logger.debug("Sending GPIO pattern to chip %s", chip)  # noqa: UP031

    chip_path = _resolve_chip_path(chip)

    line_config: dict[int | str, gpiod.LineSettings] = {
        pin: gpiod.LineSettings(
            direction=gpiod.line.Direction.OUTPUT,
            output_value=gpiod.line.Value(state),
        )
        for pin, state in pattern[0].pins.items()
    }

    with gpiod.request_lines(
        path=chip_path,
        consumer="smlight-cc-flasher",
        config=line_config,  # type: ignore[arg-type]
    ) as request:
        time.sleep(pattern[0].delay_after)

        for step in pattern[1:]:
            values: dict[int | str, gpiod.line.Value] = {
                pin: gpiod.line.Value(state) for pin, state in step.pins.items()
            }
            request.set_values(values)
            time.sleep(step.delay_after)


async def send_gpio_pattern(chip: str, pattern: list[GpioPattern]) -> None:
    """Send GPIO pattern to chip asynchronously."""
    await asyncio.get_running_loop().run_in_executor(
        None, _send_gpio_pattern, chip, pattern
    )



"""Tests for flasher module."""

from unittest.mock import AsyncMock

import pytest

from smlight_cc_flasher.flasher import Flasher


@pytest.fixture
def mock_flasher_deps(mocker):
    """Mock dependencies for Flasher class."""
    mocker.patch("smlight_cc_flasher.flasher.CommandInterface")
    mocker.patch("smlight_cc_flasher.flasher.CC26xx")
    mocker.patch("smlight_cc_flasher.flasher.Bootloader")
    mocker.patch("smlight_cc_flasher.flasher.FirmwareFile")


@pytest.mark.asyncio
async def test_flasher_init(mock_flasher_deps):
    """Test Flasher initialization."""
    flasher = Flasher("/dev/ttyUSB0")
    assert flasher._device == "/dev/ttyUSB0"
    assert flasher.ieee_address is None


@pytest.mark.asyncio
async def test_flasher_flash_with_ieee_override(mocker):
    """Test flashing with explicit IEEE address."""
    # Setup mocks
    mock_cmd = mocker.patch("smlight_cc_flasher.flasher.CommandInterface").return_value
    mock_chip = mocker.patch("smlight_cc_flasher.flasher.CC26xx").return_value
    mocker.patch("smlight_cc_flasher.flasher.Bootloader").return_value
    mocker.patch("smlight_cc_flasher.flasher.FirmwareFile")

    # Setup async methods
    mock_cmd.open = AsyncMock()
    mock_cmd.cmdReset = AsyncMock()
    mock_cmd.close = AsyncMock()
    mock_chip.connect = AsyncMock()
    mock_chip.erase = AsyncMock()
    mock_chip.flash = AsyncMock()
    mock_chip.verify = AsyncMock()
    mock_chip.set_ieee_address = AsyncMock()

    flasher = Flasher("/dev/ttyUSB0")
    # Initialize with specific address
    await flasher.async_init(ieee_address="01:02:03:04:05:06:07:08", buffer=b"dummy")

    await flasher.flash()

    # Should set parsed address
    # 0x0102030405060708
    mock_chip.set_ieee_address.assert_called_once_with(72623859790382856)
    # Ensure chip property was updated as well
    assert mock_chip.ieee_address_secondary == 72623859790382856


@pytest.mark.asyncio
async def test_flasher_flash_restore_ieee(mocker):
    """Test flashing with restoring existing IEEE address."""
    # Setup mocks
    mock_cmd = mocker.patch("smlight_cc_flasher.flasher.CommandInterface").return_value
    mock_chip = mocker.patch("smlight_cc_flasher.flasher.CC26xx").return_value
    mocker.patch("smlight_cc_flasher.flasher.Bootloader").return_value
    mocker.patch("smlight_cc_flasher.flasher.FirmwareFile")

    # Setup async methods
    mock_cmd.open = AsyncMock()
    mock_cmd.cmdReset = AsyncMock()
    mock_cmd.close = AsyncMock()
    mock_chip.connect = AsyncMock()
    mock_chip.erase = AsyncMock()
    mock_chip.flash = AsyncMock()
    mock_chip.verify = AsyncMock()
    mock_chip.set_ieee_address = AsyncMock()

    # Simulate saved address in chip (0xAABBCCDDEEFF0011)
    # 12297829382473089041
    mock_chip.ieee_address_secondary = 0xAABBCCDDEEFF0011

    flasher = Flasher("/dev/ttyUSB0")
    await flasher.async_init(buffer=b"dummy")  # No ieee_address provided

    await flasher.flash()

    mock_chip.set_ieee_address.assert_called_once_with(0xAABBCCDDEEFF0011)


@pytest.mark.asyncio
async def test_flasher_flash_no_ieee(mocker):
    """Test flashing with no IEEE address to set."""
    mock_cmd = mocker.patch("smlight_cc_flasher.flasher.CommandInterface").return_value
    mock_chip = mocker.patch("smlight_cc_flasher.flasher.CC26xx").return_value
    mocker.patch("smlight_cc_flasher.flasher.Bootloader")
    mocker.patch("smlight_cc_flasher.flasher.FirmwareFile")

    mock_cmd.open = AsyncMock()
    mock_cmd.cmdReset = AsyncMock()
    mock_cmd.close = AsyncMock()
    mock_chip.connect = AsyncMock()
    mock_chip.erase = AsyncMock()
    mock_chip.flash = AsyncMock()
    mock_chip.verify = AsyncMock()
    mock_chip.set_ieee_address = AsyncMock()

    # No saved address
    mock_chip.ieee_address_secondary = None

    flasher = Flasher("/dev/ttyUSB0")
    await flasher.async_init(buffer=b"dummy")

    await flasher.flash()

    mock_chip.set_ieee_address.assert_not_called()



# Copyright (c) 2024, SMLIGHT <smartlight.email@gmail.com>
# SPDX-License-Identifier: Apache-2.0

"""Flasher class for api use via cc-web-tools or HA"""

from collections.abc import Callable
import logging
from typing import Any

from .command import Bootloader, CommandInterface
from .device import CC26xx, parse_ieee_address
from .firmware import FirmwareFile

LOG_LEVELS = [logging.INFO, logging.DEBUG]
VERBOSE = True
_LOGGER = logging.getLogger(__name__)
logging.basicConfig(level=LOG_LEVELS[min(len(LOG_LEVELS) - 1, VERBOSE)])


class Flasher:
    command_interface: CommandInterface
    firmware: FirmwareFile | None
    chip: CC26xx
    bootloader: Bootloader
    ieee_address: str | None

    def __init__(
        self,
        device: str,
        *,
        baudrate: int = 500000,
        m33: bool = False,
        bsl2: bool = False,
    ) -> None:
        self._device = device
        self._baudrate = baudrate
        self._m33 = m33
        self.bsl2 = bsl2
        self.command_interface = CommandInterface()
        self.firmware = None
        self.ieee_address = None

    async def async_init(
        self,
        file: str | None = None,
        buffer: bytes | None = None,
        bootloader_mode: str | None = None,
        ieee_address: str | None = None,
    ) -> None:
        if file is not None and buffer is not None:
            raise ValueError("Cannot specify both file and buffer")
        if file is not None:
            self.firmware = FirmwareFile(path=file)
        elif buffer is not None:
            self.firmware = FirmwareFile(buffer=buffer)

        self.ieee_address = ieee_address
        self.chip = CC26xx(self.command_interface, self.firmware, self._m33)
        await self.command_interface.open(self._device, self._baudrate)

        self.bootloader = Bootloader(self._device, self.command_interface.transport)
        if self.bsl2:
            self.bootloader.set_mode("generic2")
        elif bootloader_mode is not None:
            self.bootloader.set_mode(bootloader_mode)

    async def connect(self) -> None:
        _LOGGER.debug("Activate bootloader")
        await self.bootloader.invoke_bootloader()
        await self.chip.connect()

    def set_firmware(self, file: FirmwareFile) -> None:
        self.chip.set_firmware(file)

    async def flash(
        self, progress_callback: Callable[[int, int], Any] | None = None
    ) -> None:
        if self.firmware is None:
            raise ValueError("Firmware not set. Initialise with file or buffer first.")
        await self.chip.erase()
        await self.chip.flash(progress_callback=progress_callback)
        await self.chip.verify()

        if self.ieee_address is not None:
            self.chip.ieee_address_secondary = parse_ieee_address(self.ieee_address)

        if self.chip.ieee_address_secondary is not None:
            await self.chip.set_ieee_address(self.chip.ieee_address_secondary)

        await self.command_interface.cmdReset()
        await self.command_interface.close()



"""Tests for firmware module."""

import binascii

import pytest

from smlight_cc_flasher.firmware import Segment


class TestSegment:
    """Tests for Segment dataclass."""

    def test_segment_creation(self):
        """Test creating a segment with valid data."""
        data = bytearray(b"\x01\x02\x03\x04\xff")
        segment = Segment(start=0x1000, end=0x1004, bytes=data)

        assert segment.start == 0x1000
        assert segment.end == 0x1004
        assert segment.size == 4
        assert len(segment.bytes) == 4
        assert segment.bytes == bytearray(b"\x01\x02\x03\x04")

    def test_segment_size_calculation(self):
        """Test that segment size is calculated correctly."""
        data = bytearray(b"\xaa\xbb\xcc\xff")
        segment = Segment(start=0x2000, end=0x2003, bytes=data)

        assert segment.size == 3
        assert segment.end - segment.start == 3

    def test_segment_crc32_calculated(self):
        """Test that CRC32 is calculated during initialization."""
        data = bytearray(b"\x01\x02\x03\x04\xff")
        segment = Segment(start=0x0, end=0x4, bytes=data)

        assert hasattr(segment, "crc32")
        assert isinstance(segment.crc32, int)
        expected_crc = binascii.crc32(bytearray(b"\x01\x02\x03\x04")) & 0xFFFFFFFF
        assert segment.crc32 == expected_crc

    def test_segment_invalid_end_byte(self):
        """Test that segment requires 0xFF as last byte."""
        data = bytearray(b"\x01\x02\x03\x04")
        with pytest.raises(AssertionError, match="Segment data does not end with 0xFF"):
            Segment(start=0x0, end=0x4, bytes=data)

    def test_segment_size_mismatch(self):
        """Test that segment validates size matches data."""
        data = bytearray(b"\x01\x02\xff")
        with pytest.raises(
            AssertionError, match="Segment size does not match data size"
        ):
            Segment(start=0x0, end=0x10, bytes=data)



class CliException(Exception):
    """CLI Exception"""


class CmdException(Exception):
    """Exception for low-level commands"""


class FwException(Exception):
    """Exception for firmware parsing"""


class DeviceException(Exception):
    """Exception for device operations"""



"""Tests for cli module."""

from unittest.mock import AsyncMock

import pytest

from smlight_cc_flasher.cli import main


@pytest.mark.asyncio
async def test_cli_ieee_override(mocker):
    """Test CLI with explicit IEEE address."""
    mocker.patch(
        "sys.argv",
        [
            "prog",
            "-d",
            "/dev/ttyUSB0",
            "-w",
            "-i",
            "01:02:03:04:05:06:07:08",
            "-f",
            "firmware.bin",
        ],
    )

    # Mock dependencies
    mock_cmd_cls = mocker.patch("smlight_cc_flasher.cli.CommandInterface")
    mock_cmd_inst = mock_cmd_cls.return_value
    mock_cmd_inst.open = AsyncMock()
    mock_cmd_inst.cmdReset = AsyncMock()
    mock_cmd_inst.close = AsyncMock()
    mock_cmd_inst.transport = mocker.Mock()

    mock_boot_cls = mocker.patch("smlight_cc_flasher.cli.Bootloader")
    mock_boot_inst = mock_boot_cls.return_value
    mock_boot_inst.invoke_bootloader = AsyncMock()

    mock_device_cls = mocker.patch("smlight_cc_flasher.cli.CC26xx")
    mock_device = mock_device_cls.return_value
    mock_device.connect = AsyncMock()
    mock_device.flash = AsyncMock()
    mock_device.set_ieee_address = AsyncMock(return_value=True)
    # Ensure erase is present since args.erase might be triggered
    mock_device.erase = AsyncMock()
    # Simulate some initial value that should be overwritten
    mock_device.ieee_address_secondary = 0xFFFFFFFFFFFFFFFF

    mock_fw_cls = mocker.patch("smlight_cc_flasher.cli.FirmwareFile")
    mock_fw_inst = mock_fw_cls.return_value
    # segments is accessed. If it returns truthy, checks loop
    mock_fw_inst.segments = []
    mock_fw_inst.check_bootloader.return_value = True
    # size is accessed for tqdm
    mock_fw_inst.size = 100

    mocker.patch("smlight_cc_flasher.cli.tqdm")

    await main()

    # Logic in CLI:
    mock_device.set_ieee_address.assert_called_once_with(72623859790382856)


@pytest.mark.asyncio
async def test_cli_ieee_restore(mocker):
    """Test CLI restoring IEEE address when not provided but found on device.
    and we are WRITING (-w).
    """
    mocker.patch("sys.argv", ["prog", "-d", "/dev/ttyUSB0", "-w", "-f", "firmware.bin"])

    mock_cmd_cls = mocker.patch("smlight_cc_flasher.cli.CommandInterface")
    mock_cmd_inst = mock_cmd_cls.return_value
    mock_cmd_inst.open = AsyncMock()
    mock_cmd_inst.cmdReset = AsyncMock()
    mock_cmd_inst.close = AsyncMock()
    mock_cmd_inst.transport = mocker.Mock()

    mock_boot_cls = mocker.patch("smlight_cc_flasher.cli.Bootloader")
    mock_boot_inst = mock_boot_cls.return_value
    mock_boot_inst.invoke_bootloader = AsyncMock()

    mock_device_cls = mocker.patch("smlight_cc_flasher.cli.CC26xx")
    mock_device = mock_device_cls.return_value
    mock_device.connect = AsyncMock()
    mock_device.flash = AsyncMock()
    mock_device.set_ieee_address = AsyncMock(return_value=True)
    mock_device.erase = AsyncMock()

    mock_fw_cls = mocker.patch("smlight_cc_flasher.cli.FirmwareFile")
    mock_fw_inst = mock_fw_cls.return_value
    mock_fw_inst.segments = []
    mock_fw_inst.check_bootloader.return_value = True
    mock_fw_inst.size = 100

    mocker.patch("smlight_cc_flasher.cli.tqdm")

    # Simulate address found on device
    mock_device.ieee_address_secondary = 0xAABBCCDDEEFF0011

    await main()

    mock_device.set_ieee_address.assert_called_once_with(0xAABBCCDDEEFF0011)


@pytest.mark.asyncio
async def test_cli_no_ieee_action_on_read(mocker):
    """Test CLI does NOT write IEEE address on read operation even if known."""
    mocker.patch(
        "sys.argv", ["prog", "-d", "/dev/ttyUSB0", "-r", "-l", "100", "-o", "dump.bin"]
    )

    mock_cmd_cls = mocker.patch("smlight_cc_flasher.cli.CommandInterface")
    mock_cmd_inst = mock_cmd_cls.return_value
    mock_cmd_inst.open = AsyncMock()
    mock_cmd_inst.cmdReset = AsyncMock()
    mock_cmd_inst.close = AsyncMock()
    mock_cmd_inst.transport = mocker.Mock()

    mock_boot_cls = mocker.patch("smlight_cc_flasher.cli.Bootloader")
    mock_boot_inst = mock_boot_cls.return_value
    mock_boot_inst.invoke_bootloader = AsyncMock()

    mock_device_cls = mocker.patch("smlight_cc_flasher.cli.CC26xx")
    mock_device = mock_device_cls.return_value
    mock_device.connect = AsyncMock()
    mock_device.read = AsyncMock()
    mock_device.set_ieee_address = AsyncMock(return_value=True)

    # Simulate address found on device
    mock_device.ieee_address_secondary = 0xAABBCCDDEEFF0011

    await main()

    # Should NOT call set_ieee_address because we are not writing and not specifying address
    mock_device.set_ieee_address.assert_not_called()



"""Tests for command module."""

import asyncio
from unittest.mock import AsyncMock, Mock

import pytest

from smlight_cc_flasher.command import Bootloader, CommandInterface, PinSetter
from smlight_cc_flasher.const import COMMAND
from smlight_cc_flasher.exceptions import CmdException


class TestPinSetter:
    """Tests for PinSetter class."""

    def test_pinsetter_initialization(self):
        """Test PinSetter initialization without webserial."""
        instance = Mock(spec=[])  # No webserial attribute
        setter = PinSetter(instance, "test_attr")

        assert setter.instance == instance
        assert setter.attr_name == "test_attr"
        assert setter._webserial is False

    def test_pinsetter_with_webserial(self):
        """Test PinSetter initialization with webserial attribute."""
        instance = Mock()
        instance.webserial = True
        instance.test_attr = False
        setter = PinSetter(instance, "test_attr")

        assert setter._webserial is True

    @pytest.mark.asyncio
    async def test_pinsetter_call_get(self):
        """Test PinSetter get value."""
        instance = Mock(spec=[])
        instance.test_attr = True
        setter = PinSetter(instance, "test_attr")

        result = await setter()
        assert result is True

    @pytest.mark.asyncio
    async def test_pinsetter_call_set(self):
        """Test PinSetter set value."""
        instance = Mock(spec=[])
        instance.test_attr = False
        setter = PinSetter(instance, "test_attr")

        result = await setter(True)
        assert instance.test_attr is True
        assert result is True

    @pytest.mark.asyncio
    async def test_pinsetter_call_webserial_set(self):
        """Test PinSetter set value with webserial."""
        instance = Mock()
        instance.webserial = True
        instance._test_attr = False
        instance.set_test_attr = AsyncMock()
        setter = PinSetter(instance, "test_attr")

        result = await setter(True)
        instance.set_test_attr.assert_called_once_with(True)
        assert result is None

    @pytest.mark.asyncio
    async def test_pinsetter_call_webserial_get(self):
        """Test PinSetter get value with webserial returns None."""
        instance = Mock()
        instance.webserial = True
        instance._test_attr = True
        setter = PinSetter(instance, "test_attr")

        result = await setter()
        assert result is None

    def test_pinsetter_get_instance(self):
        """Test PinSetter get_instance method."""
        instance = Mock()
        setter = PinSetter(instance, "test_attr")

        assert setter.get_instance() == instance


class TestBootloader:
    """Tests for Bootloader class."""

    def test_bootloader_initialization_simple(self):
        """Test Bootloader initialization with simple device."""
        transport = Mock()
        transport.serial = Mock()
        bootloader = Bootloader("/dev/ttyUSB0", transport)

        assert bootloader._device == "/dev/ttyUSB0"
        assert bootloader._serial == transport.serial
        assert bootloader._generic is True
        assert bootloader._smlight_net is False
        assert bootloader._gpio is False

    def test_bootloader_initialization_socket(self):
        """Test Bootloader initialization with socket device."""
        transport = Mock()
        transport.serial = Mock()
        bootloader = Bootloader("socket://192.168.1.1:6638", transport)

        assert bootloader._device == "socket://192.168.1.1:6638"
        assert bootloader._smlight_net is True
        assert bootloader._host == "192.168.1.1"

    def test_set_mode_generic(self):
        """Test set_mode with generic mode."""
        transport = Mock()
        transport.serial = Mock()
        bootloader = Bootloader("/dev/ttyUSB0", transport)

        bootloader.set_mode("generic")
        assert bootloader._generic is True

    def test_set_mode_generic2(self):
        """Test set_mode with generic2 mode."""
        transport = Mock()
        transport.serial = Mock()
        bootloader = Bootloader("/dev/ttyUSB0", transport)

        bootloader.set_mode("generic2")
        assert bootloader._generic is False
        assert bootloader._generic2 is True

    def test_set_mode_network(self):
        """Test set_mode with network mode."""
        transport = Mock()
        transport.serial = Mock()
        bootloader = Bootloader("/dev/ttyUSB0", transport)

        bootloader.set_mode("network", host="192.168.1.1")
        assert bootloader._generic is False
        assert bootloader._smlight_net is True
        assert bootloader._host == "192.168.1.1"

    def test_set_mode_none(self):
        """Test set_mode with none mode."""
        transport = Mock()
        transport.serial = Mock()
        bootloader = Bootloader("/dev/ttyUSB0", transport)

        bootloader.set_mode("none")
        assert bootloader._generic is False

    def test_set_mode_gpio(self):
        """Test set_mode with GPIO config name."""
        transport = Mock()
        transport.serial = Mock()
        bootloader = Bootloader("/dev/ttyUSB0", transport)

        bootloader.set_mode("smhub")
        assert bootloader._generic is False
        assert bootloader._gpio is True
        assert bootloader._gpio_config is not None

    def test_set_gpio_config_valid(self):
        """Test set_gpio_config with valid config."""
        transport = Mock()
        transport.serial = Mock()
        bootloader = Bootloader("/dev/ttyUSB0", transport)

        bootloader.set_gpio_config("smhub")
        assert bootloader._gpio_config is not None

    def test_set_gpio_config_invalid(self):
        """Test set_gpio_config with invalid config."""
        transport = Mock()
        transport.serial = Mock()
        bootloader = Bootloader("/dev/ttyUSB0", transport)

        with pytest.raises(ValueError, match="GPIO config 'invalid' not found"):
            bootloader.set_gpio_config("invalid")

    def test_set_options(self):
        """Test set_options method."""
        transport = Mock()
        transport.serial = Mock()
        bootloader = Bootloader("/dev/ttyUSB0", transport)

        bootloader.set_options(dtr_active_high=True, inverted=True)
        assert bootloader._dtr_active_high is True
        assert bootloader._inverted is True

    @pytest.mark.asyncio
    async def test_invoke_bootloader_generic(self, mocker):
        """Test invoke_bootloader with generic mode."""
        transport = Mock()
        transport.serial = Mock(spec=[])
        transport.serial.dtr = False
        transport.serial.rts = False
        bootloader = Bootloader("/dev/ttyUSB0", transport)

        mocker.patch("asyncio.sleep", new_callable=AsyncMock)

        await bootloader.invoke_bootloader()

        # Verify sleep was called
        asyncio.sleep.assert_called()

    @pytest.mark.asyncio
    async def test_invoke_bootloader_generic2(self, mocker):
        """Test invoke_bootloader with generic2 mode."""
        transport = Mock()
        transport.serial = Mock(spec=[])
        transport.serial.dtr = False
        transport.serial.rts = False
        bootloader = Bootloader("/dev/ttyUSB0", transport)
        bootloader.set_mode("generic2")

        mocker.patch("asyncio.sleep", new_callable=AsyncMock)

        await bootloader.invoke_bootloader()

        # Verify sleep was called
        asyncio.sleep.assert_called()

    @pytest.mark.asyncio
    async def test_invoke_bootloader_inverted(self, mocker):
        """Test invoke_bootloader with inverted pins."""
        transport = Mock()
        transport.serial = Mock(spec=[])
        transport.serial.dtr = False
        transport.serial.rts = False
        bootloader = Bootloader("/dev/ttyUSB0", transport)
        bootloader.set_options(dtr_active_high=False, inverted=True)

        mocker.patch("asyncio.sleep", new_callable=AsyncMock)

        await bootloader.invoke_bootloader()

        # Verify the pins were set up correctly (inverted: rts=bootloader, dtr=reset)
        assert hasattr(bootloader, "set_bootloader_pin")
        assert hasattr(bootloader, "set_reset_pin")

    @pytest.mark.asyncio
    async def test_invoke_bootloader_gpio(self, mocker):
        """Test invoke_bootloader with GPIO mode."""
        transport = Mock()
        transport.serial = Mock()
        bootloader = Bootloader("/dev/ttyUSB0", transport)
        bootloader.set_mode("smhub")

        mock_send = mocker.patch(
            "smlight_cc_flasher.gpio.send_gpio_pattern", new_callable=AsyncMock
        )
        mocker.patch("asyncio.sleep", new_callable=AsyncMock)

        await bootloader.invoke_bootloader()

        mock_send.assert_called_once()

    @pytest.mark.asyncio
    async def test_invoke_bootloader_gpio_no_config(self):
        """Test invoke_bootloader with GPIO mode but no config set."""
        transport = Mock()
        transport.serial = Mock()
        bootloader = Bootloader("/dev/ttyUSB0", transport)
        bootloader._gpio = True
        bootloader._gpio_config = None

        with pytest.raises(ValueError, match="GPIO config not set"):
            await bootloader.invoke_gpio()

    @pytest.mark.asyncio
    async def test_invoke_bootloader_network(self, mocker):
        """Test invoke_bootloader with network mode."""
        transport = Mock()
        transport.serial = Mock()
        bootloader = Bootloader("socket://192.168.1.1:6638", transport)

        # Mock the Api2 client
        mock_client = AsyncMock()
        mock_client.__aenter__ = AsyncMock(return_value=mock_client)
        mock_client.__aexit__ = AsyncMock()
        mock_client.check_auth_needed = AsyncMock(return_value=False)
        mock_client.set_cmd = AsyncMock(return_value=True)

        mock_api = mocker.patch("pysmlight.web.Api2")
        mock_api.return_value = mock_client

        mocker.patch("asyncio.sleep", new_callable=AsyncMock)

        await bootloader.invoke_bootloader()

        mock_client.check_auth_needed.assert_called_once()
        mock_client.set_cmd.assert_called_once()

    @pytest.mark.asyncio
    async def test_invoke_smlight_net_auth_needed(self, mocker):
        """Test invoke_smlight_net when auth is needed."""
        transport = Mock()
        transport.serial = Mock()
        bootloader = Bootloader("/dev/ttyUSB0", transport)

        # Mock the Api2 client
        mock_client = AsyncMock()
        mock_client.__aenter__ = AsyncMock(return_value=mock_client)
        mock_client.__aexit__ = AsyncMock()
        mock_client.check_auth_needed = AsyncMock(return_value=True)

        mock_api = mocker.patch("pysmlight.web.Api2")
        mock_api.return_value = mock_client

        result = await bootloader.invoke_smlight_net("192.168.1.1")

        assert result is False
        mock_client.check_auth_needed.assert_called_once()

    @pytest.mark.asyncio
    async def test_invoke_smlight_net_success(self, mocker):
        """Test invoke_smlight_net successful command."""
        transport = Mock()
        transport.serial = Mock()
        bootloader = Bootloader("/dev/ttyUSB0", transport)

        # Mock the Api2 client
        mock_client = AsyncMock()
        mock_client.__aenter__ = AsyncMock(return_value=mock_client)
        mock_client.__aexit__ = AsyncMock()
        mock_client.check_auth_needed = AsyncMock(return_value=False)
        mock_client.set_cmd = AsyncMock(return_value=True)

        mock_api = mocker.patch("pysmlight.web.Api2")
        mock_api.return_value = mock_client

        result = await bootloader.invoke_smlight_net("192.168.1.1")

        assert result is True
        mock_client.check_auth_needed.assert_called_once()
        mock_client.set_cmd.assert_called_once()

    @pytest.mark.asyncio
    async def test_invoke_generic_dtr_active_high(self, mocker):
        """Test invoke_generic with dtr_active_high option."""
        transport = Mock()
        transport.serial = Mock(spec=[])
        transport.serial.dtr = False
        transport.serial.rts = False
        bootloader = Bootloader("/dev/ttyUSB0", transport)
        bootloader.set_options(dtr_active_high=True, inverted=False)

        # Set up the pin setters
        bootloader.set_bootloader_pin = PinSetter(transport.serial, "dtr")
        bootloader.set_reset_pin = PinSetter(transport.serial, "rts")

        mocker.patch("asyncio.sleep", new_callable=AsyncMock)

        await bootloader.invoke_generic()

        # Verify sleep was called
        asyncio.sleep.assert_called()


class TestCommandInterface:
    """Tests for CommandInterface class."""

    @pytest.mark.asyncio
    async def test_open(self, mocker):
        """Test opening serial connection."""
        mock_reader = AsyncMock()
        mock_writer = AsyncMock()
        mock_writer.transport = Mock()
        mock_writer.transport.serial = Mock(spec=[])
        mock_writer.transport.get_extra_info = Mock(return_value=None)

        mocker.patch(
            "serial_asyncio.open_serial_connection",
            return_value=(mock_reader, mock_writer),
        )

        cmd_interface = CommandInterface()
        await cmd_interface.open("/dev/ttyUSB0", 500000)

        assert cmd_interface.reader == mock_reader
        assert cmd_interface.writer == mock_writer
        assert cmd_interface.webserial is False

    @pytest.mark.asyncio
    async def test_open_with_socket(self, mocker):
        """Test opening serial connection with TCP socket."""
        mock_reader = AsyncMock()
        mock_writer = AsyncMock()
        mock_writer.transport = Mock()
        mock_writer.transport.serial = Mock()
        mock_writer.transport.serial.webserial = False

        mock_socket = Mock()
        mock_writer.transport.get_extra_info = Mock(return_value=mock_socket)

        mocker.patch(
            "serial_asyncio.open_serial_connection",
            return_value=(mock_reader, mock_writer),
        )

        cmd_interface = CommandInterface()
        await cmd_interface.open("socket://192.168.1.1:6638", 500000)

        mock_socket.setsockopt.assert_called_once()

    @pytest.mark.asyncio
    async def test_close(self, mocker):
        """Test closing serial connection."""
        mock_reader = AsyncMock()
        mock_writer = AsyncMock()
        mock_writer.close = Mock()
        mock_writer.wait_closed = AsyncMock()
        mock_writer.transport = Mock()
        mock_writer.transport.serial = Mock()
        mock_writer.transport.serial.webserial = False
        mock_writer.transport.get_extra_info = Mock(return_value=None)

        mocker.patch(
            "serial_asyncio.open_serial_connection",
            return_value=(mock_reader, mock_writer),
        )

        cmd_interface = CommandInterface()
        await cmd_interface.open("/dev/ttyUSB0", 500000)
        await cmd_interface.close()

        mock_writer.close.assert_called_once()
        mock_writer.wait_closed.assert_called_once()

    def test_transport_property(self, mocker):
        """Test transport property."""
        cmd_interface = CommandInterface()
        cmd_interface.writer = Mock()
        cmd_interface.writer.transport = Mock()

        assert cmd_interface.transport == cmd_interface.writer.transport

    @pytest.mark.asyncio
    async def test_wait_for_ack_success(self, mocker):
        """Test waiting for ACK successfully."""
        mock_reader = AsyncMock()
        mock_writer = AsyncMock()
        mock_writer.transport = Mock()

        cmd_interface = CommandInterface()
        cmd_interface.reader = mock_reader
        cmd_interface.writer = mock_writer

        # Simulate receiving ACK
        mock_reader.read = AsyncMock(side_effect=[b"\x00", b"\xcc"])

        result = await cmd_interface._wait_for_ack(COMMAND.PING)
        assert result is True

    @pytest.mark.asyncio
    async def test_wait_for_ack_nack(self, mocker):
        """Test waiting for ACK but receiving NACK."""
        mock_reader = AsyncMock()
        mock_writer = AsyncMock()
        mock_writer.transport = Mock()

        cmd_interface = CommandInterface()
        cmd_interface.reader = mock_reader
        cmd_interface.writer = mock_writer

        # Simulate receiving NACK
        mock_reader.read = AsyncMock(side_effect=[b"\x00", b"\x33"])

        result = await cmd_interface._wait_for_ack(COMMAND.PING)
        assert result is False

    @pytest.mark.asyncio
    async def test_wait_for_ack_timeout(self, mocker):
        """Test waiting for ACK timeout."""
        mock_reader = AsyncMock()
        mock_writer = AsyncMock()
        mock_writer.transport = Mock()

        cmd_interface = CommandInterface()
        cmd_interface.reader = mock_reader
        cmd_interface.writer = mock_writer

        # Simulate timeout
        mock_reader.read = AsyncMock(side_effect=asyncio.TimeoutError)

        with pytest.raises(CmdException, match="Timeout waiting for ACK/NACK"):
            await cmd_interface._wait_for_ack(COMMAND.PING)

    @pytest.mark.asyncio
    async def test_wait_for_ack_timeout_synch(self, mocker):
        """Test waiting for ACK timeout during SYNCH command."""
        mock_reader = AsyncMock()
        mock_writer = AsyncMock()
        mock_writer.transport = Mock()

        cmd_interface = CommandInterface()
        cmd_interface.reader = mock_reader
        cmd_interface.writer = mock_writer

        # Simulate timeout during SYNCH
        mock_reader.read = AsyncMock(side_effect=asyncio.TimeoutError)

        with pytest.raises(asyncio.TimeoutError):
            await cmd_interface._wait_for_ack(COMMAND.SYNCH)

    def test_encode_addr(self):
        """Test encoding address."""
        cmd_interface = CommandInterface()
        result = cmd_interface._encode_addr(0x12345678)
        assert result == b"\x12\x34\x56\x78"

    def test_decode_addr(self):
        """Test decoding address."""
        cmd_interface = CommandInterface()
        result = cmd_interface._decode_addr(b"\x12\x34\x56\x78")
        assert result == 0x12345678

    @pytest.mark.asyncio
    async def test_write_int(self, mocker):
        """Test writing integer data."""
        mock_writer = Mock()
        mock_writer.write = Mock()
        mock_writer.drain = AsyncMock()
        mock_writer.transport = Mock()
        mock_writer.transport.get_write_buffer_size = Mock(return_value=0)

        cmd_interface = CommandInterface()
        cmd_interface.writer = mock_writer
        cmd_interface.webserial = False

        await cmd_interface._write(0x55)

        mock_writer.write.assert_called_once_with(b"\x55")
        mock_writer.drain.assert_called_once()

    @pytest.mark.asyncio
    async def test_write_bytes(self, mocker):
        """Test writing bytes data."""
        mock_writer = Mock()
        mock_writer.write = Mock()
        mock_writer.drain = AsyncMock()
        mock_writer.transport = Mock()
        mock_writer.transport.get_write_buffer_size = Mock(return_value=0)

        cmd_interface = CommandInterface()
        cmd_interface.writer = mock_writer
        cmd_interface.webserial = False

        await cmd_interface._write(b"\x01\x02\x03")

        mock_writer.write.assert_called_once_with(b"\x01\x02\x03")
        mock_writer.drain.assert_called_once()

    @pytest.mark.asyncio
    async def test_write_webserial(self, mocker):
        """Test writing with webserial."""
        mock_writer = Mock()
        mock_writer.write = Mock()
        mock_writer.drain = AsyncMock()
        mock_writer.transport = Mock()

        cmd_interface = CommandInterface()
        cmd_interface.writer = mock_writer
        cmd_interface.webserial = True

        await cmd_interface._write(b"\x01\x02\x03")

        mock_writer.write.assert_called_once_with(b"\x01\x02\x03")
        mock_writer.drain.assert_called_once()

    @pytest.mark.asyncio
    async def test_sendCmd_simple(self, mocker):
        """Test sending simple command."""
        mock_writer = Mock()
        mock_writer.write = Mock()
        mock_writer.drain = AsyncMock()
        mock_writer.transport = Mock()
        mock_writer.transport.get_write_buffer_size = Mock(return_value=0)

        cmd_interface = CommandInterface()
        cmd_interface.writer = mock_writer
        cmd_interface.webserial = False

        await cmd_interface._sendCmd(COMMAND.PING)

        mock_writer.write.assert_called_once()
        mock_writer.drain.assert_called_once()

    @pytest.mark.asyncio
    async def test_sendCmd_with_addr(self, mocker):
        """Test sending command with address."""
        mock_writer = Mock()
        mock_writer.write = Mock()
        mock_writer.drain = AsyncMock()
        mock_writer.transport = Mock()
        mock_writer.transport.get_write_buffer_size = Mock(return_value=0)

        cmd_interface = CommandInterface()
        cmd_interface.writer = mock_writer
        cmd_interface.webserial = False

        await cmd_interface._sendCmd(COMMAND.DOWNLOAD, addr=0x1000, size=0x100)

        mock_writer.write.assert_called_once()
        mock_writer.drain.assert_called_once()

    @pytest.mark.asyncio
    async def test_receivePacket_success(self, mocker):
        """Test receiving packet successfully."""
        mock_reader = AsyncMock()
        mock_writer = Mock()
        mock_writer.write = Mock()
        mock_writer.drain = AsyncMock()
        mock_writer.transport = Mock()
        mock_writer.transport.get_write_buffer_size = Mock(return_value=0)

        cmd_interface = CommandInterface()
        cmd_interface.reader = mock_reader
        cmd_interface.writer = mock_writer
        cmd_interface.webserial = False

        # Simulate packet: size=4, checksum=0x06, data=b"\x01\x02\x03"
        mock_reader.readexactly = AsyncMock(side_effect=[b"\x04\x06", b"\x01\x02\x03"])

        result = await cmd_interface.receivePacket()

        assert result == b"\x01\x02\x03"
        mock_writer.write.assert_called_once_with(CommandInterface.ACK)

    @pytest.mark.asyncio
    async def test_receivePacket_checksum_error(self, mocker):
        """Test receiving packet with checksum error."""
        mock_reader = AsyncMock()
        mock_writer = Mock()
        mock_writer.write = Mock()
        mock_writer.drain = AsyncMock()
        mock_writer.transport = Mock()
        mock_writer.transport.get_write_buffer_size = Mock(return_value=0)

        cmd_interface = CommandInterface()
        cmd_interface.reader = mock_reader
        cmd_interface.writer = mock_writer
        cmd_interface.webserial = False

        # Simulate packet with wrong checksum
        mock_reader.readexactly = AsyncMock(side_effect=[b"\x04\xff", b"\x01\x02\x03"])

        with pytest.raises(CmdException, match="checksum error"):
            await cmd_interface.receivePacket()

        mock_writer.write.assert_called_once_with(CommandInterface.NACK)

    @pytest.mark.asyncio
    async def test_cmdMemWrite_width_mismatch(self):
        """Test cmdMemWrite with width mismatch."""
        cmd_interface = CommandInterface()

        with pytest.raises(ValueError, match="width does not match len"):
            await cmd_interface.cmdMemWrite(0x1000, b"\x01\x02", 4)

    @pytest.mark.asyncio
    async def test_cmdMemWrite_invalid_width(self):
        """Test cmdMemWrite with invalid width."""
        cmd_interface = CommandInterface()

        with pytest.raises(ValueError, match="width must be 1 or 4"):
            await cmd_interface.cmdMemWrite(0x1000, b"\x01\x02", 2)

    @pytest.mark.asyncio
    async def test_cmdDownload_invalid_size(self):
        """Test cmdDownload with invalid data size."""
        cmd_interface = CommandInterface()

        with pytest.raises(CmdException, match="Invalid data size"):
            await cmd_interface.cmdDownload(0x1000, 5)

    @pytest.mark.asyncio
    async def test_cmdDownloadCRC32_invalid_size(self):
        """Test cmdDownloadCRC32 with invalid data size."""
        cmd_interface = CommandInterface()

        with pytest.raises(CmdException, match="Invalid data size"):
            await cmd_interface.cmdDownloadCRC32(0x1000, 7, 0x12345678)



# Copyright (c) 2024, SMLIGHT <smartlight.email@gmail.com>
# SPDX-License-Identifier: Apache-2.0

from enum import Enum


class COMMAND(Enum):
    PING = 0x20
    DOWNLOAD = 0x21
    GET_STATUS = 0x23
    SEND_DATA = 0x24
    RESET = 0x25
    SECTOR_ERASE = 0x26
    CRC32 = 0x27
    GET_CHIP_ID = 0x28
    MEMORY_READ = 0x2A
    MEMORY_WRITE = 0x2B
    BANK_ERASE = 0x2C
    SET_CCFG = 0x2D
    DOWNLOAD_CRC = 0x2F
    SYNCH = 0x55


class COMMAND_RET(Enum):
    SUCCESS = 0x40
    UNKNOWN_CMD = 0x41
    INVALID_CMD = 0x42
    INVALID_ADR = 0x43
    FLASH_FAIL = 0x44


COMMAND_STRS = {
    member.value: member.name.capitalize().replace("_", " ") for member in COMMAND
}

COMMAND_HAS_DATA: list[int] = [0x23, 0x27, 0x28, 0x2A]

RETURN_CMD_STRS: dict[int, str] = {
    0x40: "Success",
    0x41: "Unknown command",
    0x42: "Invalid command",
    0x43: "Invalid address",
    0x44: "Flash fail",
}

MAX_BLOCK_SIZE = 252
